'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj; };

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var pgp = require('pg-promise')();

var PostgresRelationDoesNotExistError = '42P01';
var PostgresDuplicateRelationError = '42P07';
var PostgresDuplicateColumnError = '42701';
var PostgresUniqueIndexViolationError = '23505';
var logger = require('../../../logger');

var debug = function debug() {
  var args = [].concat(Array.prototype.slice.call(arguments));
  args = ['PG: ' + arguments[0]].concat(args.slice(1, args.length));
  var log = logger.getLogger();
  log.debug.apply(log, args);
};

var parseTypeToPostgresType = function parseTypeToPostgresType(type) {
  switch (type.type) {
    case 'String':
      return 'text';
    case 'Date':
      return 'timestamp with time zone';
    case 'Object':
      return 'jsonb';
    case 'File':
      return 'jsonb';
    case 'Boolean':
      return 'boolean';
    case 'Pointer':
      return 'char(10)';
    case 'Number':
      return 'double precision';
    case 'GeoPoint':
      return 'point';
    case 'Array':
      if (type.contents && type.contents.type === 'String') {
        return 'text[]';
      } else {
        return 'jsonb';
      }
    default:
      throw 'no type for ' + JSON.stringify(type) + ' yet';
  }
};

var ParseToPosgresComparator = {
  '$gt': '>',
  '$lt': '<',
  '$gte': '>=',
  '$lte': '<='
};

var toPostgresValue = function toPostgresValue(value) {
  if ((typeof value === 'undefined' ? 'undefined' : _typeof(value)) === 'object') {
    if (value.__type === 'Date') {
      return value.iso;
    }
  }
  return value;
};

var transformValue = function transformValue(value) {
  if (value.__type == 'Pointer') {
    return value.objectId;
  }
  return value;
};

// Duplicate from then mongo adapter...
var emptyCLPS = Object.freeze({
  find: {},
  get: {},
  create: {},
  update: {},
  delete: {},
  addField: {}
});

var defaultCLPS = Object.freeze({
  find: { '*': true },
  get: { '*': true },
  create: { '*': true },
  update: { '*': true },
  delete: { '*': true },
  addField: { '*': true }
});

var toParseSchema = function toParseSchema(schema) {
  if (schema.className === '_User') {
    delete schema.fields._hashed_password;
  }
  if (schema.fields) {
    delete schema.fields._wperm;
    delete schema.fields._rperm;
  }
  var clps = defaultCLPS;
  if (schema.classLevelPermissions) {
    clps = _extends({}, emptyCLPS, schema.classLevelPermissions);
  }
  return {
    className: schema.className,
    fields: schema.fields,
    classLevelPermissions: clps
  };
};

var toPostgresSchema = function toPostgresSchema(schema) {
  if (!schema) {
    return schema;
  }
  schema.fields = schema.fields || {};
  schema.fields._wperm = { type: 'Array', contents: { type: 'String' } };
  schema.fields._rperm = { type: 'Array', contents: { type: 'String' } };
  if (schema.className === '_User') {
    schema.fields._hashed_password = { type: 'String' };
  }
  return schema;
};

var buildWhereClause = function buildWhereClause(_ref) {
  var schema = _ref.schema;
  var query = _ref.query;
  var index = _ref.index;

  var patterns = [];
  var values = [];
  var sorts = [];

  schema = toPostgresSchema(schema);

  var _loop = function _loop(fieldName) {
    var initialPatternsLength = patterns.length;
    var fieldValue = query[fieldName];
    if (fieldName.indexOf('.') >= 0) {
      var components = fieldName.split('.').map(function (cmpt, index) {
        if (index == 0) {
          return '"' + cmpt + '"';
        }
        return '\'' + cmpt + '\'';
      });
      var name = components.slice(0, components.length - 1).join('->');
      name += '->>' + components[components.length - 1];
      patterns.push(name + ' = \'' + fieldValue + '\'');
    } else if (typeof fieldValue === 'string') {
      patterns.push('$' + index + ':name = $' + (index + 1));
      values.push(fieldName, fieldValue);
      index += 2;
    } else if (typeof fieldValue === 'boolean') {
      patterns.push('$' + index + ':name = $' + (index + 1));
      values.push(fieldName, fieldValue);
      index += 2;
    } else if (typeof fieldValue === 'number') {
      patterns.push('$' + index + ':name = $' + (index + 1));
      values.push(fieldName, fieldValue);
      index += 2;
    } else if (fieldName === '$or') {
      (function () {
        var _values;

        var clauses = [];
        var clauseValues = [];
        fieldValue.forEach(function (subQuery, idx) {
          var clause = buildWhereClause({ schema: schema, query: subQuery, index: index });
          clauses.push(clause.pattern);
          clauseValues.push.apply(clauseValues, _toConsumableArray(clause.values));
          index += clause.values.length;
        });
        patterns.push('(' + clauses.join(' OR ') + ')');
        (_values = values).push.apply(_values, clauseValues);
      })();
    }

    if (fieldValue.$ne) {
      if (fieldValue.$ne === null) {
        patterns.push('$' + index + ':name <> $' + (index + 1));
      } else {
        // if not null, we need to manually exclude null
        patterns.push('($' + index + ':name <> $' + (index + 1) + ' OR $' + index + ':name IS NULL)');
      }

      // TODO: support arrays
      values.push(fieldName, fieldValue.$ne);
      index += 2;
    }

    if (fieldValue.$eq) {
      patterns.push('$' + index + ':name = $' + (index + 1));
      values.push(fieldName, fieldValue.$eq);
      index += 2;
    }
    var isInOrNin = Array.isArray(fieldValue.$in) || Array.isArray(fieldValue.$nin);
    if (Array.isArray(fieldValue.$in) && schema.fields[fieldName].type === 'Array') {
      (function () {
        var inPatterns = [];
        var allowNull = false;
        values.push(fieldName);
        fieldValue.$in.forEach(function (listElem, listIndex) {
          if (listElem === null) {
            allowNull = true;
          } else {
            values.push(listElem);
            inPatterns.push('$' + (index + 1 + listIndex - (allowNull ? 1 : 0)));
          }
        });
        if (allowNull) {
          patterns.push('($' + index + ':name IS NULL OR $' + index + ':name && ARRAY[' + inPatterns.join(',') + '])');
        } else {
          patterns.push('$' + index + ':name && ARRAY[' + inPatterns.join(',') + ']');
        }
        index = index + 1 + inPatterns.length;
      })();
    } else if (isInOrNin) {
      createConstraint = function createConstraint(baseArray, notIn) {
        if (baseArray.length > 0) {
          (function () {
            var inPatterns = [];
            values.push(fieldName);
            baseArray.forEach(function (listElem, listIndex) {
              values.push(listElem);
              inPatterns.push('$' + (index + 1 + listIndex));
            });
            var not = notIn ? 'NOT' : '';
            patterns.push('$' + index + ':name ' + not + ' IN (' + inPatterns.join(',') + ')');
            index = index + 1 + inPatterns.length;
          })();
        } else if (!notIn) {
          values.push(fieldName);
          patterns.push('$' + index + ':name IS NULL');
          index = index + 1;
        }
      };

      if (fieldValue.$in) {
        createConstraint(fieldValue.$in, false);
      }
      if (fieldValue.$nin) {
        createConstraint(fieldValue.$nin, true);
      }
    }

    if (typeof fieldValue.$exists !== 'undefined') {
      if (fieldValue.$exists) {
        patterns.push('$' + index + ':name IS NOT NULL');
      } else {
        patterns.push('$' + index + ':name IS NULL');
      }
      values.push(fieldName);
      index += 1;
    }

    if (fieldValue.$nearSphere) {
      var point = fieldValue.$nearSphere;
      var distance = fieldValue.$maxDistance;
      var distanceInKM = distance * 6371 * 1000;
      patterns.push('ST_distance_sphere($' + index + ':name::geometry, POINT($' + (index + 1) + ', $' + (index + 2) + ')::geometry) <= $' + (index + 3));
      sorts.push('ST_distance_sphere($' + index + ':name::geometry, POINT($' + (index + 1) + ', $' + (index + 2) + ')::geometry) ASC');
      values.push(fieldName, point.latitude, point.longitude, distanceInKM);
      index += 4;
    }

    if (fieldValue.$regex) {
      var regex = fieldValue.$regex;
      var operator = '~';
      var opts = fieldValue.$options;
      if (opts) {
        if (opts.indexOf('i') >= 0) {
          operator = '~*';
        }
      }
      patterns.push('$' + index + ':name ' + operator + ' $' + (index + 1));
      values.push(fieldName, regex);
      index += 2;
    }

    if (fieldValue.__type === 'Pointer') {
      patterns.push('$' + index + ':name = $' + (index + 1));
      values.push(fieldName, fieldValue.objectId);
      index += 2;
    }

    if (fieldValue.__type === 'Date') {
      patterns.push('$' + index + ':name = $' + (index + 1));
      values.push(fieldName, fieldValue.iso);
      index += 2;
    }

    Object.keys(ParseToPosgresComparator).forEach(function (cmp) {
      if (fieldValue[cmp]) {
        var pgComparator = ParseToPosgresComparator[cmp];
        patterns.push('$' + index + ':name ' + pgComparator + ' $' + (index + 1));
        values.push(fieldName, toPostgresValue(fieldValue[cmp]));
        index += 2;
      }
    });

    if (initialPatternsLength === patterns.length) {
      throw new Parse.Error(Parse.Error.OPERATION_FORBIDDEN, 'Postgres doesn\'t support this query type yet ' + JSON.stringify(fieldValue));
    }
  };

  for (var fieldName in query) {
    var createConstraint;

    _loop(fieldName);
  }
  values = values.map(transformValue);
  return { pattern: patterns.join(' AND '), values: values, sorts: sorts };
};

var PostgresStorageAdapter = exports.PostgresStorageAdapter = function () {
  // Private
  function PostgresStorageAdapter(_ref2) {
    var uri = _ref2.uri;
    var _ref2$collectionPrefi = _ref2.collectionPrefix;
    var collectionPrefix = _ref2$collectionPrefi === undefined ? '' : _ref2$collectionPrefi;

    _classCallCheck(this, PostgresStorageAdapter);

    this._collectionPrefix = collectionPrefix;
    this._client = pgp(uri);
  }

  _createClass(PostgresStorageAdapter, [{
    key: '_ensureSchemaCollectionExists',
    value: function _ensureSchemaCollectionExists() {
      return this._client.none('CREATE TABLE "_SCHEMA" ( "className" varChar(120), "schema" jsonb, "isParseClass" bool, PRIMARY KEY ("className") )').catch(function (error) {
        if (error.code === PostgresDuplicateRelationError || error.code === PostgresUniqueIndexViolationError) {
          // Table already exists, must have been created by a different request. Ignore error.
        } else {
          throw error;
        }
      });
    }
  }, {
    key: 'classExists',
    value: function classExists(name) {
      return notImplemented();
    }
  }, {
    key: 'setClassLevelPermissions',
    value: function setClassLevelPermissions(className, CLPs) {
      var _this = this;

      return this._ensureSchemaCollectionExists().then(function () {
        var values = [className, 'schema', 'classLevelPermissions', CLPs];
        return _this._client.none('UPDATE "_SCHEMA" SET $2:name = json_object_set_key($2:name, $3::text, $4::jsonb) WHERE "className"=$1 ', values);
      }).catch(function (err) {
        console.error("ERR!!!", err);
        return Promise.reject(err);
      });
    }
  }, {
    key: 'createClass',
    value: function createClass(className, schema) {
      var _this2 = this;

      return this.createTable(className, schema).then(function () {
        return _this2._client.none('INSERT INTO "_SCHEMA" ("className", "schema", "isParseClass") VALUES ($<className>, $<schema>, true)', { className: className, schema: schema });
      }).then(function () {
        return toParseSchema(schema);
      });
    }

    // Just create a table, do not insert in schema

  }, {
    key: 'createTable',
    value: function createTable(className, schema) {
      var _this3 = this;

      debug('createTable', className, schema);
      var valuesArray = [];
      var patternsArray = [];
      var fields = Object.assign({}, schema.fields);
      if (className === '_User') {
        fields._email_verify_token_expires_at = { type: 'Date' };
        fields._email_verify_token = { type: 'String' };
      }
      var index = 2;
      var relations = [];
      Object.keys(fields).forEach(function (fieldName) {
        var parseType = fields[fieldName];
        // Skip when it's a relation
        // We'll create the tables later
        if (parseType.type == 'Relation') {
          relations.push(fieldName);
          return;
        }
        if (['_rperm', '_wperm'].includes(fieldName)) {
          parseType.contents = { type: 'String' };
        }
        valuesArray.push(fieldName);
        valuesArray.push(parseTypeToPostgresType(parseType));
        patternsArray.push('$' + index + ':name $' + (index + 1) + ':raw');
        index = index + 2;
      });
      var qs = 'CREATE TABLE $1:name (' + patternsArray.join(',') + ', PRIMARY KEY ("objectId"))';
      var values = [className].concat(valuesArray);
      return this._ensureSchemaCollectionExists().then(function () {
        return _this3._client.none(qs, values);
      }).catch(function (error) {
        if (error.code === PostgresDuplicateRelationError) {
          // Table already exists, must have been created by a different request. Ignore error.
        } else {
          throw error;
        }
      }).then(function () {
        // Create the relation tables
        return Promise.all(relations.map(function (fieldName) {
          return _this3._client.none('CREATE TABLE IF NOT EXISTS $<joinTable:name> ("relatedId" varChar(120), "owningId" varChar(120), PRIMARY KEY("relatedId", "owningId") )', { joinTable: '_Join:' + fieldName + ':' + className });
        }));
      });
    }
  }, {
    key: 'addFieldIfNotExists',
    value: function addFieldIfNotExists(className, fieldName, type) {
      var _this4 = this;

      // TODO: Must be revised for invalid logic...
      debug('addFieldIfNotExists', { className: className, fieldName: fieldName, type: type });
      return this._client.tx("addFieldIfNotExists", function (t) {
        var promise = Promise.resolve();
        if (type.type !== 'Relation') {
          promise = t.none('ALTER TABLE $<className:name> ADD COLUMN $<fieldName:name> $<postgresType:raw>', {
            className: className,
            fieldName: fieldName,
            postgresType: parseTypeToPostgresType(type)
          }).catch(function (error) {
            if (error.code === PostgresRelationDoesNotExistError) {
              return _this4.createClass(className, { fields: _defineProperty({}, fieldName, type) });
            } else if (error.code === PostgresDuplicateColumnError) {
              // Column already exists, created by other request. Carry on to
              // See if it's the right type.
            } else {
              throw error;
            }
          });
        } else {
          promise = t.none('CREATE TABLE IF NOT EXISTS $<joinTable:name> ("relatedId" varChar(120), "owningId" varChar(120), PRIMARY KEY("relatedId", "owningId") )', { joinTable: '_Join:' + fieldName + ':' + className });
        }
        return promise.then(function () {
          return t.any('SELECT "schema" FROM "_SCHEMA" WHERE "className" = $<className>', { className: className });
        }).then(function (result) {
          if (fieldName in result[0].schema) {
            throw "Attempted to add a field that already exists";
          } else {
            result[0].schema.fields[fieldName] = type;
            return t.none('UPDATE "_SCHEMA" SET "schema"=$<schema> WHERE "className"=$<className>', { schema: result[0].schema, className: className });
          }
        });
      });
    }

    // Drops a collection. Resolves with true if it was a Parse Schema (eg. _User, Custom, etc.)
    // and resolves with false if it wasn't (eg. a join table). Rejects if deletion was impossible.

  }, {
    key: 'deleteClass',
    value: function deleteClass(className) {
      return notImplemented();
    }

    // Delete all data known to this adapter. Used for testing.

  }, {
    key: 'deleteAllClasses',
    value: function deleteAllClasses() {
      var _this5 = this;

      var now = new Date().getTime();
      debug('deleteAllClasses');
      return this._client.any('SELECT * FROM "_SCHEMA"').then(function (results) {
        var joins = results.reduce(function (list, schema) {
          Object.keys(schema.schema.fields).forEach(function (field) {
            if (schema.schema.fields[field].type === 'Relation') {
              list.push('_Join:' + field + ':' + schema.className);
            }
          });
          return list;
        }, []);
        var classes = ['_SCHEMA', '_PushStatus', '_Hooks', '_GlobalConfig'].concat(_toConsumableArray(results.map(function (result) {
          return result.className;
        })), _toConsumableArray(joins));
        return _this5._client.tx(function (t) {
          return t.batch(classes.map(function (className) {
            return t.none('DROP TABLE IF EXISTS $<className:name>', { className: className });
          }));
        });
      }, function (error) {
        if (error.code === PostgresRelationDoesNotExistError) {
          // No _SCHEMA collection. Don't delete anything.
          return;
        } else {
          throw error;
        }
      }).then(function () {
        debug('deleteAllClasses done in ' + (new Date().getTime() - now));
      });
    }

    // Remove the column and all the data. For Relations, the _Join collection is handled
    // specially, this function does not delete _Join columns. It should, however, indicate
    // that the relation fields does not exist anymore. In mongo, this means removing it from
    // the _SCHEMA collection.  There should be no actual data in the collection under the same name
    // as the relation column, so it's fine to attempt to delete it. If the fields listed to be
    // deleted do not exist, this function should return successfully anyways. Checking for
    // attempts to delete non-existent fields is the responsibility of Parse Server.

    // This function is not obligated to delete fields atomically. It is given the field
    // names in a list so that databases that are capable of deleting fields atomically
    // may do so.

    // Returns a Promise.

  }, {
    key: 'deleteFields',
    value: function deleteFields(className, schema, fieldNames) {
      return notImplemented();
    }

    // Return a promise for all schemas known to this adapter, in Parse format. In case the
    // schemas cannot be retrieved, returns a promise that rejects. Requirements for the
    // rejection reason are TBD.

  }, {
    key: 'getAllClasses',
    value: function getAllClasses() {
      var _this6 = this;

      return this._ensureSchemaCollectionExists().then(function () {
        return _this6._client.map('SELECT * FROM "_SCHEMA"', null, function (row) {
          return _extends({ className: row.className }, row.schema);
        });
      }).then(function (res) {
        return res.map(toParseSchema);
      });
    }

    // Return a promise for the schema with the given name, in Parse format. If
    // this adapter doesn't know about the schema, return a promise that rejects with
    // undefined as the reason.

  }, {
    key: 'getClass',
    value: function getClass(className) {
      return this._client.any('SELECT * FROM "_SCHEMA" WHERE "className"=$<className>', { className: className }).then(function (result) {
        if (result.length === 1) {
          return result[0].schema;
        } else {
          throw undefined;
        }
      }).then(toParseSchema);
    }

    // TODO: remove the mongo format dependency in the return value

  }, {
    key: 'createObject',
    value: function createObject(className, schema, object) {
      debug('createObject', className, object);
      var columnsArray = [];
      var newFieldsArray = [];
      var valuesArray = [];
      schema = toPostgresSchema(schema);
      var geoPoints = {};
      Object.keys(object).forEach(function (fieldName) {
        var authDataMatch = fieldName.match(/^_auth_data_([a-zA-Z0-9_]+)$/);
        if (authDataMatch) {
          var provider = authDataMatch[1];
          object['authData'] = object['authData'] || {};
          object['authData'][provider] = object[fieldName];
          delete object[fieldName];
          fieldName = 'authData';
        }

        columnsArray.push(fieldName);
        if (!schema.fields[fieldName] && className === '_User') {
          if (fieldName == '_email_verify_token') {
            valuesArray.push(object[fieldName]);
          }
          if (fieldName == '_email_verify_token_expires_at') {
            valuesArray.push(object[fieldName].iso);
          }
          return;
        }
        switch (schema.fields[fieldName].type) {
          case 'Date':
            valuesArray.push(object[fieldName].iso);
            break;
          case 'Pointer':
            valuesArray.push(object[fieldName].objectId);
            break;
          case 'Array':
            if (['_rperm', '_wperm'].includes(fieldName)) {
              valuesArray.push(object[fieldName]);
            } else {
              valuesArray.push(JSON.stringify(object[fieldName]));
            }
            break;
          case 'Object':
          case 'String':
          case 'Number':
          case 'Boolean':
          case 'File':
            valuesArray.push(object[fieldName]);
            break;
          case 'GeoPoint':
            // pop the point and process later
            geoPoints[fieldName] = object[fieldName];
            columnsArray.pop();
            break;
          default:
            throw 'Type ' + schema.fields[fieldName].type + ' not supported yet';
            break;
        }
      });

      columnsArray = columnsArray.concat(Object.keys(geoPoints));
      var initialValues = valuesArray.map(function (val, index) {
        return '$' + (index + 2 + columnsArray.length) + (['_rperm', '_wperm'].includes(columnsArray[index]) ? '::text[]' : '');
      });

      var geoPointsInjects = Object.keys(geoPoints).map(function (key, idx) {
        var value = geoPoints[key];
        valuesArray.push(value.latitude, value.longitude);
        var l = valuesArray.length + columnsArray.length;
        return 'POINT($' + l + ', $' + (l + 1) + ')';
      });

      var columnsPattern = columnsArray.map(function (col, index) {
        return '$' + (index + 2) + ':name';
      }).join(',');
      var valuesPattern = initialValues.concat(geoPointsInjects).join(',');

      var qs = 'INSERT INTO $1:name (' + columnsPattern + ') VALUES (' + valuesPattern + ')';
      var values = [className].concat(_toConsumableArray(columnsArray), valuesArray);
      debug(qs, values);
      return this._client.any(qs, values).then(function () {
        return { ops: [object] };
      }).catch(function (error) {
        if (error.code === PostgresUniqueIndexViolationError) {
          throw new Parse.Error(Parse.Error.DUPLICATE_VALUE, 'A duplicate value for a field with unique values was provided');
        } else {
          throw error;
        }
      });
    }

    // Remove all objects that match the given Parse Query.
    // If no objects match, reject with OBJECT_NOT_FOUND. If objects are found and deleted, resolve with undefined.
    // If there is some other error, reject with INTERNAL_SERVER_ERROR.

  }, {
    key: 'deleteObjectsByQuery',
    value: function deleteObjectsByQuery(className, schema, query) {
      debug('deleteObjectsByQuery', className, query);
      var values = [className];
      var index = 2;
      var where = buildWhereClause({ schema: schema, index: index, query: query });
      values.push.apply(values, _toConsumableArray(where.values));
      if (Object.keys(query).length == 0) {
        where.pattern = 'TRUE';
      }
      var qs = 'WITH deleted AS (DELETE FROM $1:name WHERE ' + where.pattern + ' RETURNING *) SELECT count(*) FROM deleted';
      debug(qs, values);
      return this._client.one(qs, values, function (a) {
        return +a.count;
      }).then(function (count) {
        if (count === 0) {
          throw new Parse.Error(Parse.Error.OBJECT_NOT_FOUND, 'Object not found.');
        } else {
          return count;
        }
      });
    }

    // Apply the update to all objects that match the given Parse Query.

  }, {
    key: 'updateObjectsByQuery',
    value: function updateObjectsByQuery(className, schema, query, update) {
      debug('updateObjectsByQuery', className, query, update);
      return this.findOneAndUpdate(className, schema, query, update);
    }

    // Return value not currently well specified.

  }, {
    key: 'findOneAndUpdate',
    value: function findOneAndUpdate(className, schema, query, update) {
      debug('findOneAndUpdate', className, query, update);
      var conditionPatterns = [];
      var updatePatterns = [];
      var values = [className];
      var index = 2;
      schema = toPostgresSchema(schema);
      for (var fieldName in update) {
        var _fieldValue = update[fieldName];
        var authDataMatch = fieldName.match(/^_auth_data_([a-zA-Z0-9_]+)$/);
        if (authDataMatch) {
          var provider = authDataMatch[1];
          var value = update[fieldName];
          delete update[fieldName];
          fieldName = 'authData';
          updatePatterns.push('$' + index + ':name = json_object_set_key($' + index + ':name, $' + (index + 1) + '::text, $' + (index + 2) + '::jsonb)');
          values.push(fieldName, provider, value);
          index += 3;
        } else if (_fieldValue.__op === 'Increment') {
          updatePatterns.push('$' + index + ':name = COALESCE($' + index + ':name, 0) + $' + (index + 1));
          values.push(fieldName, _fieldValue.amount);
          index += 2;
        } else if (_fieldValue.__op === 'Add') {
          updatePatterns.push('$' + index + ':name = COALESCE($' + index + ':name, \'[]\'::jsonb) || $' + (index + 1));
          values.push(fieldName, _fieldValue.objects);
          index += 2;
        } else if (_fieldValue.__op === 'Delete') {
          updatePatterns.push('$' + index + ':name = $' + (index + 1));
          values.push(fieldName, null);
          index += 2;
        } else if (_fieldValue.__op === 'Remove') {
          return Promise.reject(new Parse.Error(Parse.Error.OPERATION_FORBIDDEN, 'Postgres does not support Remove operator.'));
        } else if (_fieldValue.__op === 'AddUnique') {
          return Promise.reject(new Parse.Error(Parse.Error.OPERATION_FORBIDDEN, 'Postgres does not support AddUnique operator.'));
        } else if (fieldName === 'updatedAt') {
          //TODO: stop special casing this. It should check for __type === 'Date' and use .iso
          updatePatterns.push('$' + index + ':name = $' + (index + 1));
          values.push(fieldName, _fieldValue);
          index += 2;
        } else if (typeof _fieldValue === 'string') {
          updatePatterns.push('$' + index + ':name = $' + (index + 1));
          values.push(fieldName, _fieldValue);
          index += 2;
        } else if (typeof _fieldValue === 'boolean') {
          updatePatterns.push('$' + index + ':name = $' + (index + 1));
          values.push(fieldName, _fieldValue);
          index += 2;
        } else if (_fieldValue.__type === 'Pointer') {
          updatePatterns.push('$' + index + ':name = $' + (index + 1));
          values.push(fieldName, _fieldValue.objectId);
          index += 2;
        } else if (_fieldValue.__type === 'Date') {
          updatePatterns.push('$' + index + ':name = $' + (index + 1));
          values.push(fieldName, toPostgresValue(_fieldValue));
          index += 2;
        } else if (_fieldValue.__type === 'GeoPoint') {
          updatePatterns.push('$' + index + ':name = POINT($' + (index + 1) + ', $' + (index + 2) + ')');
          values.push(fieldName, _fieldValue.latitude, _fieldValue.longitude);
          index += 3;
        } else if (typeof _fieldValue === 'number') {
          updatePatterns.push('$' + index + ':name = $' + (index + 1));
          values.push(fieldName, _fieldValue);
          index += 2;
        } else if ((typeof _fieldValue === 'undefined' ? 'undefined' : _typeof(_fieldValue)) === 'object' && schema.fields[fieldName] && schema.fields[fieldName].type == 'Object') {
          updatePatterns.push('$' + index + ':name = $' + (index + 1));
          values.push(fieldName, _fieldValue);
          index += 2;
        } else if (Array.isArray(_fieldValue) && schema.fields[fieldName] && schema.fields[fieldName].type == 'Array') {
          var expectedType = parseTypeToPostgresType(schema.fields[fieldName]);
          if (expectedType === 'text[]') {
            updatePatterns.push('$' + index + ':name = $' + (index + 1) + '::text[]');
          } else {
            updatePatterns.push('$' + index + ':name = array_to_json($' + (index + 1) + '::text[])::jsonb');
          }
          values.push(fieldName, _fieldValue);
          index += 2;
        } else {
          debug('Not supported update', fieldName, _fieldValue);
          return Promise.reject(new Parse.Error(Parse.Error.OPERATION_FORBIDDEN, 'Postgres doesn\'t support update ' + JSON.stringify(_fieldValue) + ' yet'));
        }
      }

      var where = buildWhereClause({ schema: schema, index: index, query: query });
      values.push.apply(values, _toConsumableArray(where.values));

      var qs = 'UPDATE $1:name SET ' + updatePatterns.join(',') + ' WHERE ' + where.pattern + ' RETURNING *';
      debug('update: ', qs, values);
      return this._client.any(qs, values).then(function (val) {
        return val[0];
      }); // TODO: This is unsafe, verification is needed, or a different query method;
    }

    // Hopefully, we can get rid of this. It's only used for config and hooks.

  }, {
    key: 'upsertOneObject',
    value: function upsertOneObject(className, schema, query, update) {
      debug('upsertOneObject', { className: className, query: query, update: update });
      return this.createObject(className, schema, update).catch(function (err) {
        // ignore duplicate value errors as it's upsert
        if (err.code == Parse.Error.DUPLICATE_VALUE) {
          return;
        }
        throw err;
      });
    }
  }, {
    key: 'find',
    value: function find(className, schema, query, _ref3) {
      var skip = _ref3.skip;
      var limit = _ref3.limit;
      var sort = _ref3.sort;

      debug('find', className, query, { skip: skip, limit: limit, sort: sort });
      var hasLimit = limit !== undefined;
      var hasSkip = skip !== undefined;
      var values = [className];
      var where = buildWhereClause({ schema: schema, query: query, index: 2 });
      values.push.apply(values, _toConsumableArray(where.values));

      var wherePattern = where.pattern.length > 0 ? 'WHERE ' + where.pattern : '';
      var limitPattern = hasLimit ? 'LIMIT $' + (values.length + 1) : '';
      if (hasLimit) {
        values.push(limit);
      }
      var skipPattern = hasSkip ? 'OFFSET $' + (values.length + 1) : '';
      if (hasSkip) {
        values.push(skip);
      }

      var sortPattern = '';
      if (sort) {
        var sorting = Object.keys(sort).map(function (key) {
          // Using $idx pattern gives:  non-integer constant in ORDER BY
          if (sort[key] === 1) {
            return '"' + key + '" ASC';
          }
          return '"' + key + '" DESC';
        }).join(',');
        sortPattern = sort !== undefined && Object.keys(sort).length > 0 ? 'ORDER BY ' + sorting : '';
      }
      if (where.sorts && Object.keys(where.sorts).length > 0) {
        sortPattern = 'ORDER BY ' + where.sorts.join(',');
      }

      var qs = 'SELECT * FROM $1:name ' + wherePattern + ' ' + sortPattern + ' ' + limitPattern + ' ' + skipPattern;
      debug(qs, values);
      return this._client.any(qs, values).catch(function (err) {
        // Query on non existing table, don't crash
        if (err.code === PostgresRelationDoesNotExistError) {
          return [];
        }
        return Promise.reject(err);
      }).then(function (results) {
        return results.map(function (object) {
          Object.keys(schema.fields).forEach(function (fieldName) {
            if (schema.fields[fieldName].type === 'Pointer' && object[fieldName]) {
              object[fieldName] = { objectId: object[fieldName], __type: 'Pointer', className: schema.fields[fieldName].targetClass };
            }
            if (schema.fields[fieldName].type === 'Relation') {
              object[fieldName] = {
                __type: "Relation",
                className: schema.fields[fieldName].targetClass
              };
            }
            if (object[fieldName] && schema.fields[fieldName].type === 'GeoPoint') {
              object[fieldName] = {
                latitude: object[fieldName].x,
                longitude: object[fieldName].y
              };
            }
          });
          //TODO: remove this reliance on the mongo format. DB adapter shouldn't know there is a difference between created at and any other date field.
          if (object.createdAt) {
            object.createdAt = object.createdAt.toISOString();
          }
          if (object.updatedAt) {
            object.updatedAt = object.updatedAt.toISOString();
          }
          if (object.expiresAt) {
            object.expiresAt = { __type: 'Date', iso: object.expiresAt.toISOString() };
          }
          if (object._email_verify_token_expires_at) {
            object._email_verify_token_expires_at = { __type: 'Date', iso: object._email_verify_token_expires_at.toISOString() };
          }

          for (var fieldName in object) {
            if (object[fieldName] === null) {
              delete object[fieldName];
            }
            if (object[fieldName] instanceof Date) {
              object[fieldName] = { __type: 'Date', iso: object[fieldName].toISOString() };
            }
          }

          return object;
        });
      });
    }

    // Create a unique index. Unique indexes on nullable fields are not allowed. Since we don't
    // currently know which fields are nullable and which aren't, we ignore that criteria.
    // As such, we shouldn't expose this function to users of parse until we have an out-of-band
    // Way of determining if a field is nullable. Undefined doesn't count against uniqueness,
    // which is why we use sparse indexes.

  }, {
    key: 'ensureUniqueness',
    value: function ensureUniqueness(className, schema, fieldNames) {
      // Use the same name for every ensureUniqueness attempt, because postgres
      // Will happily create the same index with multiple names.
      var constraintName = 'unique_' + fieldNames.sort().join('_');
      var constraintPatterns = fieldNames.map(function (fieldName, index) {
        return '$' + (index + 3) + ':name';
      });
      var qs = 'ALTER TABLE $1:name ADD CONSTRAINT $2:name UNIQUE (' + constraintPatterns.join(',') + ')';
      return this._client.none(qs, [className, constraintName].concat(_toConsumableArray(fieldNames))).catch(function (error) {
        if (error.code === PostgresDuplicateRelationError && error.message.includes(constraintName)) {
          // Index already exists. Ignore error.
        } else if (error.code === PostgresUniqueIndexViolationError && error.message.includes(constraintName)) {
          // Cast the error into the proper parse error
          throw new Parse.Error(Parse.Error.DUPLICATE_VALUE, 'A duplicate value for a field with unique values was provided');
        } else {
          throw error;
        }
      });
    }

    // Executes a count.

  }, {
    key: 'count',
    value: function count(className, schema, query) {
      var values = [className];
      var where = buildWhereClause({ schema: schema, query: query, index: 2 });
      values.push.apply(values, _toConsumableArray(where.values));

      var wherePattern = where.pattern.length > 0 ? 'WHERE ' + where.pattern : '';
      var qs = 'SELECT count(*) FROM $1:name ' + wherePattern;
      return this._client.one(qs, values, function (a) {
        return +a.count;
      });
    }
  }, {
    key: 'performInitialization',
    value: function performInitialization(_ref4) {
      var _this7 = this;

      var VolatileClassesSchemas = _ref4.VolatileClassesSchemas;

      var now = new Date().getTime();
      debug('performInitialization');
      var promises = VolatileClassesSchemas.map(function (schema) {
        return _this7.createTable(schema.className, schema);
      });

      return Promise.all(promises).then(function () {
        return _this7._client.any(json_object_set_key).catch(function (err) {
          console.error(err);
        });
      }).then(function () {
        debug('initialzationDone in ' + (new Date().getTime() - now));
      });
    }
  }]);

  return PostgresStorageAdapter;
}();

function notImplemented() {
  return Promise.reject(new Error('Not implemented yet.'));
}

// Function to set a key on a nested JSON document
var json_object_set_key = 'CREATE OR REPLACE FUNCTION "json_object_set_key"(\
  "json"          jsonb,\
  "key_to_set"    TEXT,\
  "value_to_set"  anyelement\
)\
  RETURNS jsonb \
  LANGUAGE sql \
  IMMUTABLE \
  STRICT \
AS $function$\
SELECT concat(\'{\', string_agg(to_json("key") || \':\' || "value", \',\'), \'}\')::jsonb\
  FROM (SELECT *\
          FROM jsonb_each("json")\
         WHERE "key" <> "key_to_set"\
         UNION ALL\
        SELECT "key_to_set", to_json("value_to_set")::jsonb) AS "fields"\
$function$;';

exports.default = PostgresStorageAdapter;

module.exports = PostgresStorageAdapter; // Required for tests